/**********************************************************/
/*  apimagic: cparser-based API normalization utility     */
/*  Copyright (C) 2015--2021  SysDeer Technologies, LLC   */
/*  Released under GPLv2 and GPLv3; see COPYING.APIMAGIC. */
/**********************************************************/

#include <stdio.h>
#include <unistd.h>
#include <apimagic/apimagic.h>
#include "apimagic_driver_impl.h"
#include "apimagic_dprintf_impl.h"

#ifndef AMGC_DRIVER_FLAGS
#define AMGC_DRIVER_FLAGS	AMGC_DRIVER_VERBOSITY_ERRORS \
				| AMGC_DRIVER_VERBOSITY_USAGE
#endif

static const char vermsg[] = "%s%s%s (git://midipix.org/apimagic): "
			     "version %s%d.%d.%d%s.\n"
			     "[commit reference: %s%s%s]\n";

static const char * const amgc_ver_color[6] = {
		"\x1b[1m\x1b[35m","\x1b[0m",
		"\x1b[1m\x1b[32m","\x1b[0m",
		"\x1b[1m\x1b[34m","\x1b[0m"
};

static const char * const amgc_ver_plain[6] = {
		"","",
		"","",
		"",""
};

static ssize_t amgc_version(struct amgc_driver_ctx * dctx, int fdout)
{
	const struct amgc_source_version * verinfo;
	const char * const * verclr;

	verinfo = amgc_source_version();
	verclr  = isatty(fdout) ? amgc_ver_color : amgc_ver_plain;

	return amgc_dprintf(
			fdout,vermsg,
			verclr[0],dctx->program,verclr[1],
			verclr[2],verinfo->major,verinfo->minor,
			verinfo->revision,verclr[3],
			verclr[4],verinfo->commit,verclr[5]);
}

static void amgc_perform_unit_actions(
	const struct amgc_driver_ctx *	dctx,
	struct amgc_unit_ctx *		uctx)
{
	const struct amgc_action * action;

	for (action=uctx->cctx->actions; action->type; action++)
		amgc_perform_unit_action(dctx,uctx,action,0);
}

static int amgc_exit(struct amgc_driver_ctx * dctx, int ret)
{
	amgc_output_error_vector(dctx);
	amgc_free_driver_ctx(dctx);
	return ret;
}

int amgc_main(char ** argv, char ** envp, const struct amgc_fd_ctx * fdctx)
{
	int				ret;
	int				fdout;
	uint64_t			flags;
	struct amgc_driver_ctx *	dctx;
	struct amgc_unit_ctx *		uctx;
	const char **			unit;

	flags = AMGC_DRIVER_FLAGS;
	fdout = fdctx ? fdctx->fdout : STDOUT_FILENO;

	if ((ret = amgc_get_driver_ctx(argv,envp,flags,fdctx,&dctx)))
		return (ret == AMGC_USAGE)
			? !argv || !argv[0] || !argv[1]
			: AMGC_ERROR;

	if (dctx->cctx->drvflags & AMGC_DRIVER_VERSION)
		if ((amgc_version(dctx,fdout)) < 0)
			return amgc_exit(dctx,AMGC_ERROR);

	for (unit=dctx->units; *unit && !dctx->errv[0]; unit++) {
		if (!(amgc_get_unit_ctx(dctx,*unit,&uctx))) {
			amgc_perform_unit_actions(dctx,uctx);
			amgc_free_unit_ctx(uctx);
		}
	}

	return amgc_exit(dctx,dctx->errv[0] ? AMGC_ERROR : AMGC_OK);
}
